 /*
Program Description:

This program reads in a data file and generates a trace of dictionary 
operations that can be used by the DIMACS Dictionary driver. 
The input file parameters are as follows:
    ini X            number of initial inserts 
    rep Y            number of repetitions
    ins              include inserts in repetitions
    suc              include successful lookups in repetitions
    uns              include unsuccessful lookups
    del              include deletes
    sed R            optional random seed
    key K            integer keyparts between 0..K-1

All of these have default values (see dict.h).  

The program generates X inserts with random keys, followed by Y
repetitions of whichever operations in [ins suc uns del] are specified 
by user. 

The insert keys are formed by random integers from 0..K-1, with a 
decimal timestamp added to ensure unique keys.  Timestamps are all
below .5, and decrease with time.  Unsuccessful lookups are formed by 
chosing a valid key in the dictionary at random, and changing it to 
have an invalid timestamp.  

Written by Marc Brooks with some modifications by C. McGeoch  8/96

*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include "dict.h"

cmdtype cmdtable[CMD_SIZE+1];    /* table of user commands               */
int namer;                       /* for unique key timestamps */ 

/*----table_init---------------------------------------------------------*/
void table_init ( ) {
/*
Function Description:
This is simple an initialization function. It is used in the parse() 
function.
*/

  strcpy(cmdtable[0], "com");
  strcpy(cmdtable[1], "ins");   /* whether to use insert in loop */
  strcpy(cmdtable[2], "suc");   /* whether to do successful lookups */
  strcpy(cmdtable[3], "uns");   /* whether to do unsuccess  lookups */
  strcpy(cmdtable[4], "del");   /* whether to do deletes  */ 
  strcpy(cmdtable[5], "rep");    /* how many reps in loop */
  strcpy(cmdtable[6], "ini");    /* how many initial inserts */ 
  strcpy(cmdtable[7], "sed");    /* optional seed input */ 
  strcpy(cmdtable[8], "key");    /* optional max key value */ 

} /* end of table_init */

/*-----param_init--------------------------------------------------------*/ 
void param_init(Pparameters param) {

  param->num_ops = default_num_ops; 
  param->init_ins = default_init_ins;
  param->ins = default_ins;
  param->suc_lookup = default_suc_lookup;
  param->unsuc_lookup = default_unsuc_lookup;
  param->del = default_del;
  param->myseed = (int) time (0);
  param->maxkey = (double) KEY_RANGE;

} /* end of param_init */

/*------parse-----------------------------------------------------------*/
int parse( char* cmd ) {
/*
Function Description:
This function takes a command and return a integer index. 
*/

  int i;
  int go = 1;

  for (i = CMD_SIZE ; go != 0; i--) {
    go = strncmp(cmdtable[i], cmd,3 ); 
  }
  return i+1;
}/*lookup*/

/*---read_paramters---------------------------------------------------*/ 
void read_parameters ( Pparameters param) {
/*

Function Description:
This function reads through the input and writes the user supplied 
parameters to the param.

*/

  cmdtype cmd;
  char tmp[100];
  int n; 

  while ( scanf("%s", cmd) != EOF ) {
    fgets ( tmp, sizeof(tmp), stdin) ; 
    switch ( parse( cmd ) ) {
    case 0 : { /* comment */
	       break;
	     } /* end of case 0 */
    case 1 : { /* use insert */ 
               param->ins = 1;
	       break;
	     } /* end of case 1 */
    case 2 : { /* do successful */
	       param->suc_lookup = 1;
	       break;
	     } /* end of case 2 */
    case 3 : { /* do unsuccessful */
               param->unsuc_lookup = 1;
	       break;
	     } /* end of case 3 */
    case 4 : { /* do deletes */
               param->del = 1;
	       break;
	     } /* end of case 4 */
    case 5 : {/*number repetitions */
              sscanf (tmp, "%d", &n);
              param->num_ops = n; 
              break;
	    } /* end of case 5*/
    case 6 : {/*number of initial inserts*/
              sscanf (tmp, "%d", &n );
              param->init_ins = n;
              break;
	    }/* end of case 6*/
    case 7 : {/* user-supplied seed */
             sscanf (tmp, "%d", &n);
             param -> myseed = n;
             break;
           } /*end of case 7 */
    case 8 : {/* max intpart of key */
             sscanf (tmp, "%d", &n);
             param->maxkey = (double) n;
             break;
           } /* end of case 8 */
    } /* end of switch statement */
  } /* end of while loop */

} /* end of read_parameters */

/*---write_param--------------------------------------------------*/ 
void write_param( Pparameters param ) {

  printf(  "dch 0\n"); /* header line */ 
  printf(  "com Dictionary operations with random keys. \n");
  printf(  "com Initial Inserts = %d, repetitions = %d\n",
            param->init_ins, param->num_ops); 
  printf(  "com Insert? %d\n", param->ins);
  printf(  "com Successful Lookup? %d\n", param->suc_lookup);
  printf(  "com Unsuccessful Lookup? %d\n", param->unsuc_lookup);
  printf(  "com Delete? %d\n", param->del); 
  printf(  "com Seed = %d\n", param->myseed); 
  printf(  "com Int keypart range 0.. %.0lf  \n", param->maxkey-1); 

} /* end of print_param */

/*---init_dict------------------------------------------------------*/ 
void init_dict ( Pkey_stor key_holder, Pparameters param ) {
/* initial inserts in dictionary */

  int i;

  for ( i = 0; i <= key_holder->last_key; i++ ) {
       printf("ins %lf \n", key_holder->keys[i] );
  } /* end of for loop */

} /* end of init_dict */

/*-----generate_key---------------------------------------------------*/ 
double generate_key( int i, Pparameters param ) {
/* Valid keys always have digital part less than or equal to .5      */
  int temp;

  temp = param->maxkey * drand48();   /* gives 0...KEY_RANGE-1 */ 
                                      /* then add digital timestamp part*/ 
  return( (double) temp + 1.0 / ((double) i + 2.0));
} 

/*-----generate_init_keys---------------------------------------------*/ 
void generate_init_keys ( Pkey_stor key_holder, Pparameters param ) {

  int i;
  srand48(param->myseed); 

  for ( i = 0; i < param->init_ins ; i++ ) {
        key_holder->keys[i] = generate_key(i, param);
  } /* end of for loop */
  key_holder->last_key = param->init_ins -1;
  namer = param->init_ins; 

} /* end of generate_init_keys */

/*---write_ins--------------------------------------------------------*/ 
void write_ins ( Pkey_stor key_holder, Pparameters param ) {

  key_holder->keys[key_holder->last_key+1] = 
     generate_key( namer++, param); 
  key_holder->last_key++;
  printf ( "ins %lf\n", key_holder->keys[key_holder->last_key] );
 
} /* end of write_ins */

/*---write_suc_lookup------------------------------------------------*/
void write_suc_lookup ( Pkey_stor key_holder ) {
/* Chooses a random key from list of dictionary contents.            */


  printf ( "lkp %lf\n", 
	  (key_holder->keys[(int)((key_holder->last_key) * 
					  drand48())]) );

} /* end write_suc_lookup */

/*---write_unsuc_lookup---------------------------------------------*/ 

void write_unsuc_lookup ( Pkey_stor key_holder, Pparameters param ) {
/* Generates a random key and adds an impossible digital timestamp. */ 

int tmp;
    /* get a valid one and make an impossible timestamp */ 
    
    tmp = (int) (key_holder->last_key * drand48() );
    printf("lkp %lf\n", key_holder->keys[tmp]  + .49);

} /* end of write_unsuc_lookup */

/*---key_shuffle------------------------------------------------------*/ 
void key_shuffle ( int doomed_key, int last_key, Pkey_stor key_holder ) {

  key_holder->keys[doomed_key] = key_holder->keys[last_key];
  key_holder->last_key -= 1;

} /* end of swap */

/*---write_del---------------------------------------------------------*/
void write_del ( Pkey_stor key_holder ) {

  int doomed_key;
  
  doomed_key = (key_holder->last_key+1) * drand48();
  printf( "dlk %lf\n", key_holder->keys[doomed_key] );
  key_shuffle( doomed_key, key_holder->last_key, key_holder );

} /* end of write_del */

/*---write_outfile---------------------------------------------------*/
void write_outfile ( Pparameters param ) {

  Pkey_stor key_holder;
  int inserted[param->num_ops];
  int i;

  key_holder = (Pkey_stor)malloc(sizeof(key_stor));
  generate_init_keys( key_holder , param);

  write_param( param );

  init_dict( key_holder, param );

  for ( i = 0; i < param->num_ops; i++ ) {
    if ( param->ins )        write_ins ( key_holder, param );
    if ( param->suc_lookup ) write_suc_lookup (  key_holder );
    if ( param->unsuc_lookup)write_unsuc_lookup ( key_holder, param );
    if ( param->del )        write_del ( key_holder );
  } /* end of for loop */


} /* end of write_outfile */

/*---main-------------------------------------------------*/
void main( int argc,        /* Number of strings in array argv          */
          char *argv[] )    /* Array of command-line argument strings   */
{

  parameters *param;
    
  table_init();
  param = (Pparameters)malloc(sizeof(parameters));
  param_init( param );

  read_parameters( param );
  write_outfile( param );

  free (param); 
} /* end of main */

